import pandas as pd
import numpy as np
import datetime
import os
import re

import utils

class ConductCalibProvider:
    used_channels = ["C1", "C2"]

    def __init__(self):
        self.std_dir = dict()
        self.calib_runs = dict()

    def open_calibration_run(self, directory: str):
        if not os.path.isdir(directory):
            raise ValueError("Input parameter is not a directory")

        calib_run = CalibRun(directory)
        self.calib_runs[calib_run.date_str+calib_run.time_str] = calib_run

        return calib_run

    def export_set_as_file(self, filename: str):
        set_df = pd.DataFrame()
        for run_id in self.calib_runs:
            calib_run = self.calib_runs[run_id]
            run_df = calib_run.steps_df.copy()
            run_df["standard_cond"] = calib_run.standard_cond
            run_df["run_id"] = run_id
            set_df = set_df.append(run_df)

        set_df.to_csv(path_or_buf=filename,
                      sep="\t",
                      index=False)


class CalibRun:

    VALID_STD_THRESHOLD = 0.01

    def __init__(self, directory_path: str):
        # Date and time
        self.date_str = ""
        self.time_str = ""
        self.update_datetime_from_dir(directory_path)

        # Real conductivity value of the standard
        self.standard_cond = 0.0

        # Data dataframes
        self.periodic_df = pd.DataFrame()
        self.instant_config_df = pd.DataFrame()
        self.user_config_df = pd.DataFrame()
        self.steps_df = pd.DataFrame()

        self.analyse_files(directory_path)

        self.update_steps_df()

    def update_steps_df(self):
        config_df = self.instant_config_df.copy()

        # From long to wide format
        config_df = config_df.pivot(index='datetime', columns='name', values='value')

        # Forward fill NAs. Values not changing are not recorded in instant_config_df, so fill with previous value.
        config_df = config_df.fillna(method='ffill')

        # Compute step number for each line
        config_df["step"] = np.arange(len(config_df)) + 1

        # Set start/end datetime for each step
        config_df["start_datetime"] = config_df.index
        config_df["end_datetime"] = config_df["start_datetime"].shift(-1)

        # For the last step, compute the end_datetime from expected length written in User-defined config file.
        config_df.loc[config_df.index[-1], 'end_datetime'] = \
            config_df.loc[config_df.index[-1], 'start_datetime'] \
            + datetime.timedelta(seconds=int(self.user_config_df.loc[self.user_config_df.index[-1], 'length_sec']))
        config_df.reset_index(drop=True, inplace=True)

        # Reshape: separate column for excite and headgain
        config_df = pd.wide_to_long(config_df, ["excite", "headgain"], i="step", j="channel", sep="_").reset_index()

        # Add "stable" timeframe limit for each step, i.e. remove transition phases at the beginning/end of each step.
        config_df["stable_start_datetime"] = config_df["start_datetime"] + datetime.timedelta(seconds=2)
        config_df["stable_end_datetime"] = config_df["end_datetime"] - datetime.timedelta(seconds=2)

        # Compute mean, standard deviation and validity for each step/channelComment
        for channel in range(1, 6):
            channel_config_df = config_df[config_df["channel"] == channel].copy()
            for index, row in channel_config_df.iterrows():
                subset_df = self.periodic_df[(self.periodic_df["datetime"] >= row["stable_start_datetime"])
                                             & (self.periodic_df["datetime"] < row["stable_end_datetime"])]
                channel_config_df.loc[index, "mean"] = subset_df["C" + str(channel)].mean()
                channel_config_df.loc[index, "std"] = subset_df["C" + str(channel)].std()
                channel_config_df.loc[index, "channel"] = channel
                channel_config_df.loc[index, "valid_auto"] = channel_config_df.loc[index, "std"] < self.VALID_STD_THRESHOLD
                channel_config_df.loc[index, "valid_manu"] = channel_config_df.loc[index, "valid_auto"]
            self.steps_df = self.steps_df.append(channel_config_df)

    def update_datetime_from_dir(self, directory: str) -> None:
        folders = directory.split("/")
        last_subfolder_parts = folders[len(folders)-1].split("_")
        self.date_str = last_subfolder_parts[0]
        self.time_str = last_subfolder_parts[1]

    def analyse_files(self, directory: str):
        files = os.listdir(directory)
        directory += "/"

        # Periodic
        periodic_file = self.__find_single_file_in_file_list__(
            files_list=files,
            file_pattern=r"[0-9]{8}_[0-9]{6}_conduct-calib_(EDAQ|CONDUCTI)_periodic.log",
            file_identifyer="Periodic")
        self.periodic_df = self.get_periodic_file_content(directory + periodic_file)

        # Instant configuration changes
        instant_config_file = self.__find_single_file_in_file_list__(
            files_list=files,
            file_pattern=r"[0-9]{8}_[0-9]{6}_conduct-calib_(EDAQ|CONDUCTI)_instant.log",
            file_identifyer="Instant config")
        self.instant_config_df = self.get_instant_config_file_content(directory + instant_config_file)

        # User-defined) configuration changes
        user_config_file = self.__find_single_file_in_file_list__(
            files_list=files,
            file_pattern=r"conduct_calib.csv",
            file_identifyer="User config")
        self.user_config_df = self.get_user_config_file_content(directory + user_config_file)

        # Compare expected and real configuration changes.
        config_ok = self.__check_real_config__(self.instant_config_df, self.user_config_df)
        # if not config_ok:
        #     raise ValueError("Real configuration file does not match user-defined config.")

        # Real conductivity value of the standard
        standard_cond_file = self.__find_single_file_in_file_list__(
            files_list=files,
            file_pattern=r"standard_conductivity.txt",
            file_identifyer="Standard conductivity")
        with open(directory + standard_cond_file, "r") as file:
            self.standard_cond = float(file.readline())

    def __check_real_config__(self, real_config_df: pd.DataFrame, user_config_df: pd.DataFrame) -> bool:
        """Compare the user-defined configuration file with the instant recording of the configuration changes.

        Note: Currently only check steps number"""
        return real_config_df["step"].max() == user_config_df["step"].max()

    def __find_single_file_in_file_list__(self, files_list: list, file_pattern: str, file_identifyer: str) -> str:
        found_files = [f for f in files_list if re.search(file_pattern, f)]
        if len(found_files) > 1:
            raise ValueError("More than one file of type [" + file_identifyer + "] found.")
        elif len(found_files) == 0:
            raise ValueError("No file found for type [" + file_identifyer + "]")

        return found_files[0]

    def get_periodic_file_content(self, filename: str) -> pd.DataFrame:
        periodic_df = pd.read_csv(filename, sep="\t", parse_dates=["datetime"])
        return periodic_df

    def get_instant_config_file_content(self, filename: str) -> pd.DataFrame:
        config_df = pd.read_csv(filename, sep=",", parse_dates=["datetime"])

        # Identify steps.
        # Configuration changes related to a given step should occur within few milliseconds, so use the datetime of
        # configuration changes to detect steps.
        config_df["step"] = config_df["datetime"].diff().gt(datetime.timedelta(seconds=1)).cumsum() + 1

        # Use the datetime of the first configuration change as the common datetime for all the changes of each step.
        config_df["datetime"] = config_df.groupby("step")["datetime"].transform('min')

        return config_df

    def get_user_config_file_content(self, filename: str) -> pd.DataFrame:
        config_df = pd.read_csv(filename, sep=",")
        return config_df