import pyqtgraph as pg
from PyQt5.QtWidgets import *
from PyQt5.QtCore import *
import pandas as pd
import numpy as np

import utils
from dataprovider.conductcalibprovider import ConductCalibProvider
from dataprovider.conductcalibprovider import CalibRun
from gui.uimainwindow import Ui_MainWindow


class ConductCalibUim:

    # Calibration "set" tableWidget columns identifiers.
    DATE_COL = 0
    TIMESTART_COL = 1
    STANDARDCOND_COL = 2
    STEPS_COL = 3

    # Calibration "run" tableWidget columns identifiers.
    MEAN_COL = 0
    STD_COL = 1
    FREQ_COL = 2
    EXCITE_COL = 3
    HEADGAIN_COL = 4
    VALIDAUTO_COL = 5
    VALIDMANUAL_COL = 6

    def __init__(self, conduct_prvd: ConductCalibProvider, main_ui: Ui_MainWindow):
        self.main_ui = main_ui
        self.conduct_prvd = conduct_prvd
        self.current_calib_run = None
        self.current_channel = int(self.main_ui.conduct_combobox_channel.currentText())

        self.__initialize_set_table__()
        self.__initialize_run_plot__()
        self.__initialize_run_table__()

        self.main_ui.conduct_pushbutton_load_run.clicked.connect(self.load_data)
        self.main_ui.conduct_pushbutton_save_csv.clicked.connect(self.save_as_csv)

        self.main_ui.conduct_tablewidget_set.cellClicked.connect(self.change_selected_run)
        self.main_ui.conduct_combobox_channel.currentTextChanged.connect(self.change_selected_channel)

        self.main_ui.conduct_tablewidget_run.cellClicked.connect(self.highlight_step)

    def load_data(self):
        directory = QFileDialog.getExistingDirectory()

        # If user cancels directory selection, do nothing.
        if directory == "":
            return

        # Try to open/analyse calibration run, display a error popup if failure.
        try:
            calib_run = self.conduct_prvd.open_calibration_run(directory)
        except ValueError as e:
            msg = QMessageBox()
            msg.setIcon(QMessageBox.Critical)
            msg.setWindowTitle("Failed to open calibration run.")
            msg.setText(str(e))
            msg.exec_()
            return

        self.current_calib_run = calib_run
        self.__add_new_row_in_set_table__(calib_run)
        self.__update_run_plot__(calib_run, self.current_channel)
        self.__update_run_table__(calib_run, self.current_channel)

    def save_as_csv(self):
        filename = QFileDialog.getSaveFileName(caption="Save Calibration set as CSV...", filter="*.csv")
        if filename[0] != '':
            self.conduct_prvd.export_set_as_file(filename[0])

    ####################################################################################################################
    # Calibration set table

    def __initialize_set_table__(self):
        """Initialize the set table"""
        # Select whole row (instead of selecting cell)
        self.main_ui.conduct_tablewidget_set.setSelectionBehavior(QAbstractItemView.SelectRows)

        # Set column widths
        self.main_ui.conduct_tablewidget_set.setColumnWidth(self.DATE_COL, 80)
        self.main_ui.conduct_tablewidget_set.setColumnWidth(self.TIMESTART_COL, 75)
        self.main_ui.conduct_tablewidget_set.setColumnWidth(self.STANDARDCOND_COL, 95)
        self.main_ui.conduct_tablewidget_set.setColumnWidth(self.STEPS_COL, 80)

    def __add_new_row_in_set_table__(self, calib_run: CalibRun):
        """Add a new row in the calibration table. Initialize the cell's content and properties."""
        row_id = self.main_ui.conduct_tablewidget_set.rowCount()
        self.main_ui.conduct_tablewidget_set.insertRow(row_id)

        # Date
        date_item = QTableWidgetItem()
        date_item.setFlags(Qt.ItemIsEnabled)  # Read only
        date_item.setText(calib_run.date_str)
        self.main_ui.conduct_tablewidget_set.setItem(row_id, self.DATE_COL, date_item)

        # Time
        time_item = QTableWidgetItem()
        time_item.setFlags(Qt.ItemIsEnabled)  # Read only
        time_item.setText(calib_run.time_str)
        self.main_ui.conduct_tablewidget_set.setItem(row_id, self.TIMESTART_COL, time_item)

        # Standard conductivity
        standard_cond_item = QTableWidgetItem()
        standard_cond_item.setFlags(Qt.ItemIsEnabled)  # Read only
        standard_cond_item.setText(str(calib_run.standard_cond))
        self.main_ui.conduct_tablewidget_set.setItem(row_id, self.STANDARDCOND_COL, standard_cond_item)

        # Number of steps in the run
        steps_item = QTableWidgetItem()
        steps_item.setFlags(Qt.ItemIsEnabled)  # Read only
        steps_item.setText(str(calib_run.instant_config_df["step"].max()))
        self.main_ui.conduct_tablewidget_set.setItem(row_id, self.STEPS_COL, steps_item)

        utils.highlight_row(self.main_ui.conduct_tablewidget_set, row_id)

    def change_selected_run(self, row: int, column: int):
        # Get run identifier (date and time concatenated)
        identifier = self.main_ui.conduct_tablewidget_set.item(row, self.DATE_COL).text() \
                     + self.main_ui.conduct_tablewidget_set.item(row, self.TIMESTART_COL).text()

        # Get calib run
        calib_run = self.conduct_prvd.calib_runs[identifier]

        self.current_calib_run = calib_run

        utils.highlight_row(self.main_ui.conduct_tablewidget_set, row)

        self.__update_run_plot__(calib_run, self.current_channel)
        self.__update_run_table__(calib_run, self.current_channel)

    def change_selected_channel(self, channel: str):
        self.current_channel = int(channel)
        self.__update_run_plot__(self.current_calib_run, int(channel))
        self.__update_run_table__(self.current_calib_run, int(channel))

    ####################################################################
    # "Run" table

    def __initialize_run_table__(self):
        """Initialize the run table"""
        # Select whole row (instead of selecting cell)
        self.main_ui.conduct_tablewidget_run.setSelectionBehavior(QAbstractItemView.SelectRows)

        # Set column widths
        self.main_ui.conduct_tablewidget_run.setColumnWidth(self.MEAN_COL, 55)
        self.main_ui.conduct_tablewidget_run.setColumnWidth(self.STD_COL, 55)
        self.main_ui.conduct_tablewidget_run.setColumnWidth(self.FREQ_COL, 40)
        self.main_ui.conduct_tablewidget_run.setColumnWidth(self.HEADGAIN_COL, 40)
        self.main_ui.conduct_tablewidget_run.setColumnWidth(self.EXCITE_COL, 40)
        self.main_ui.conduct_tablewidget_run.setColumnWidth(self.VALIDAUTO_COL, 50)
        self.main_ui.conduct_tablewidget_run.setColumnWidth(self.VALIDMANUAL_COL, 50)

    def __update_run_table__(self, calib_run: CalibRun, channel: int):
        # Clear table
        self.main_ui.conduct_tablewidget_run.setRowCount(0)

        dataset_df = calib_run.steps_df[calib_run.steps_df["channel"] == channel]
        for index, row in dataset_df.iterrows():
            self.__add_new_row_in_run_table__(row)

    def __add_new_row_in_run_table__(self, row_data: pd.Series):
        """Add a new row in the calibration table. Initialize the cell's content and properties."""
        row_id = self.main_ui.conduct_tablewidget_run.rowCount()
        self.main_ui.conduct_tablewidget_run.insertRow(row_id)

        # Mean
        mean_item = QTableWidgetItem()
        mean_item.setFlags(Qt.ItemIsEnabled)  # Read only
        mean_item.setText("{:.4f}".format(row_data["mean"]))
        self.main_ui.conduct_tablewidget_run.setItem(row_id, self.MEAN_COL, mean_item)

        # Std
        std_item = QTableWidgetItem()
        std_item.setFlags(Qt.ItemIsEnabled)  # Read only
        std_item.setText("{:.4f}".format(row_data["std"]))
        self.main_ui.conduct_tablewidget_run.setItem(row_id, self.STD_COL, std_item)

        # Frequency
        frequency_item = QTableWidgetItem()
        frequency_item.setFlags(Qt.ItemIsEnabled)  # Read only
        frequency_item.setText(str(row_data["frequency"]))
        self.main_ui.conduct_tablewidget_run.setItem(row_id, self.FREQ_COL, frequency_item)

        # Excite
        excite_item = QTableWidgetItem()
        excite_item.setFlags(Qt.ItemIsEnabled)  # Read only
        excite_item.setText(str(row_data["excite"]))
        self.main_ui.conduct_tablewidget_run.setItem(row_id, self.EXCITE_COL, excite_item)

        # Headgain
        headgain_item = QTableWidgetItem()
        headgain_item.setFlags(Qt.ItemIsEnabled)  # Read only
        headgain_item.setText(str(row_data["headgain"]))
        self.main_ui.conduct_tablewidget_run.setItem(row_id, self.HEADGAIN_COL, headgain_item)

        # Valid automatic
        validauto_item = QCheckBox()
        validauto_item.setEnabled(False)  # Read only
        validauto_item.setChecked(row_data["valid_auto"])
        self.main_ui.conduct_tablewidget_run.setCellWidget(row_id, self.VALIDAUTO_COL, validauto_item)

        # Valid manually
        validmanu_item = QCheckBox()
        validmanu_item.setEnabled(True)
        validmanu_item.setChecked(row_data["valid_manu"])
        self.main_ui.conduct_tablewidget_run.setCellWidget(row_id, self.VALIDMANUAL_COL, validmanu_item)
        self.main_ui.conduct_tablewidget_run.cellWidget(row_id, self.VALIDMANUAL_COL).stateChanged.connect(
            lambda checkstate, row_id=row_id: self.__change_step_validity__(step_id=row_id+1, validity=checkstate))

    def __change_step_validity__(self, step_id: int, validity: bool):
        self.current_calib_run.steps_df.loc[(self.current_calib_run.steps_df["step"] == step_id)
                                            & (self.current_calib_run.steps_df["channel"] == self.current_channel),
                                            "valid_manu"] = bool(validity)
        self.__update_run_plot__(self.current_calib_run, self.current_channel)

    ####################################################################
    # "Run" plot

    def __initialize_run_plot__(self) -> None:

        self.plot_item = pg.PlotItem(axisItems={'bottom': utils.TimeAxisItem(orientation='bottom')},
                                     title='Conductivity')

        self.curve = pg.PlotCurveItem()
        self.plot_item.addItem(self.curve)

        self.step_points = pg.ScatterPlotItem(size=10, pen=pg.mkPen('w'))
        self.plot_item.addItem(self.step_points)

        self.main_ui.conduct_graphicsview_run.setCentralItem(self.plot_item)

        self.lines_items = []

    def __update_run_plot__(self, calib_run: CalibRun, channel: int) -> None:
        # Remove previous vertical lines
        for line_item in self.lines_items:
            self.plot_item.removeItem(line_item)

        # Vertical line for configuration change
        config_df = calib_run.instant_config_df.copy()
        config_pen = pg.mkPen(color=(200, 200, 255), style=Qt.DotLine)
        for instant in utils.datetime64_to_epoch_ms(config_df["datetime"].unique()):
            vertical_line = pg.InfiniteLine(pos=instant, pen=config_pen)
            self.plot_item.addItem(vertical_line)
            self.lines_items.append(vertical_line)

        # Conductivity points
        periodic_df = calib_run.periodic_df.copy()
        periodic_df = periodic_df[periodic_df["datetime"] >= min(config_df["datetime"])]

        self.curve.setData(x=utils.pd_time_to_epoch_ms(periodic_df['datetime']),
                           y=list(periodic_df["C" + str(channel)]),
                           pen=(200, 200, 200),
                           symbol="o",
                           symbolBrush=(255, 0, 0),
                           symbolPen='w')

        # Step average points
        self.set_step_points_data(calib_run, channel)

    def set_step_points_data(self, calib_run: CalibRun, channel: int, highlight_index: int=None):
        steps_df = calib_run.steps_df[calib_run.steps_df["channel"] == channel]
        steps_df.reset_index(drop=True, inplace=True)

        avg_time = np.average(np.array([utils.pd_time_to_epoch_ms(steps_df["stable_start_datetime"]),
                                        utils.pd_time_to_epoch_ms(steps_df["stable_end_datetime"])]),
                              axis=0)

        # Symbol brush for highlighting
        # symbol_brushs = [[0, 255, 0]] * len(steps_df.index)
        symbol_brushs = [[0, 255, 0] if row["valid_manu"] else [255, 0, 0] for index, row in steps_df.iterrows()]
        if highlight_index is not None:
            symbol_brushs[highlight_index] = [0, 0, 255]

        self.step_points.clear()
        spots = [{'pos': (avg_time[step_id],
                          steps_df.at[step_id, "mean"]),
                  'data': step_id,
                  'brush': symbol_brushs[step_id]}
                 for step_id in range(len(steps_df.index))]
        self.step_points.addPoints(spots)

    def highlight_step(self, row: int, column: int):
        self.set_step_points_data(self.current_calib_run, self.current_channel, highlight_index=row)



